/*  C O N N E C T (w/ minitelnet)
 *
 * Resolves an internet name into an IP number, and then connects to it.
 * (like telnet, but raw)
 *
 * By Brian Gaeke <brg@dgate.org>
 * (C) 1994, 1995 The Dimensional Gate Company. All rights reserved.
 * See file `Copyright' for your rights concerning this software package.
 */

static char rcsid[] = "$Id: connect.c,v 1.1 1995/07/18 20:50:22 brg Exp $";

#include <config.h>

#include <sys/types.h>
#include <sys/signal.h>
#if HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#if HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <stdio.h>

#if !HAVE_SELECT
# error "Cannot compile connect without select() function!"
#endif /* !HAVE_SELECT */
#if !HAVE_SOCKET
# error "Cannot compile connect without socket() function!"
#endif /* !HAVE_SOCKET */

int silent = 0;
int telnet = 0;
int telnet_echo = 1;			/* state saved from telnet IAC DO/DONT ECHO */
int telnet_stopnow = 0;			/* tells minitelnet to stop now (broken...) */

unsigned long resolve(char *name)
{
	struct hostent *h;

	h = gethostbyname(name);
	if (!h)
		return 0xFFFFFFFF;
	return ((struct in_addr *) h->h_addr)->s_addr;
}

int get_socket(void)
{
	int s;

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
	{
		perror("Error: socket");
		exit(1);
	}
	return s;
}

RETSIGTYPE telnet_sigint(int sig)
{
	telnet_stopnow++;
}

void DoIt(int s)
{
	fd_set readfds;
	int rc, len, x;
	unsigned char buf[1024];

	while (1)
	{
		if (telnet && telnet_stopnow)
		{
			close(s);
			printf("Connection terminated.\n");
			return;
		}
		FD_ZERO(&readfds);
		FD_SET(0, &readfds);
		FD_SET(s, &readfds);
		rc = select(s + 1, &readfds, (fd_set *) 0, (fd_set *) 0,
			(struct timeval *) 0);
		if (rc == -1)
		{
			perror("select");
			continue;
		}
		if (rc == 0)
			continue;
		if (FD_ISSET(0, &readfds))
		{
			len = read(0, buf, 1024);
			if (len < 0)
			{
				perror("Error reading from stdin");
				return;
			}
			else if (len == 0)
			{
				shutdown(s, 2);
				if (!silent)
					printf("Connection terminated.\n");
				return;
			}
			if (write(s, buf, len) != len)
			{
				perror("Error writing to network");
				return;
			}
		}
		if (FD_ISSET(s, &readfds))
		{
			len = read(s, buf, 1024);
			if (len < 0)
			{
				perror("Error reading from network");
				return;
			}
			else if (len == 0)
			{
				if (!silent)
					printf("Connection closed by foreign host.\n");
				return;
			}
			if (telnet)
				for (x = 0; x < 1024; x++)
				{
					/* fake-out responses to Interpret As Command sequences */
					if (buf[x] == IAC)
					{
						if (buf[x + 1] == DO || buf[x + 1] == DONT)
						{
							unsigned char telcmd[16];
							unsigned char insist = buf[x + 1];
							unsigned char whatever = buf[x + 2];
							unsigned char reply;

							/* there's framework here for more processing... 
							 * currently we just say we WONT do anything
							 */
							reply = WONT;
							sprintf(telcmd, "%c%c%c", IAC, reply, whatever);
							if (write(s, telcmd, 3) != 3)
							{
								perror("Error writing to network");
								return;
							}
							buf[x + 2] = 0;
						}
						buf[x] = 0;
						buf[x + 1] = 0;
					}
				}
			if ((telnet && telnet_echo) || !telnet)
				if (write(1, buf, len) != len)
				{
					perror("Error writing to stdout");
					return;
				}
		}
	}
}

void usage(char *myname)
{
	fprintf(stderr,
			"  Usage: %s hostname portnumber [-st]\n"
			"  Common ports: telnet = 23, smtp = 25, http = 80.\n\n"
			"    -s  --  be silent about status & error messages\n"
			"    -t  --  we are talking to a telnetd; act accordingly\n\n",
			myname);
	exit(2);
}

main(int argc, char **argv)
{
	struct sockaddr_in addr;
	unsigned long theaddr;
	int rc, port, s, x, y;
	char *hostname;

	if (argc < 3)
		usage(argv[0]);
	hostname = argv[1];
	port = atoi(argv[2]);
	for (x = 3; x < argc; x++)
	{
		if (argv[x][0] == '-')
			for (y=1;y<strlen(argv[x]);y++)
			switch (argv[x][y])
			{
				case 's':
					silent++;
					break;
				case 't':
					telnet++;
					if (!silent)
						printf("Doing telnet protocol fake-out.\n");
					break;
				default:
					usage(argv[0]);
					break;
			}
		else
			usage(argv[0]);
	}

	if (!silent)
		printf("Trying %s ", hostname);
	(void) fflush(stdout);
	theaddr = resolve(hostname);
	if (theaddr == 0xFFFFFFFF)
	{
		fprintf(stderr,
				"\n\n%s: Error resolving hostname (name server error or"
				" nonexistent host.)\n",
				argv[0]);
		exit(3);
	}
	s = get_socket();

	/* fill in an address */
	memset(&addr, 0, sizeof(struct sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = theaddr;
	addr.sin_port = htons(port);

	if (!silent)
		printf("(%s), port %u...", inet_ntoa(addr.sin_addr), port);
	(void) fflush(stdout);
	/* connect the socket */
	if (connect(s, (struct sockaddr *) &addr, sizeof(struct sockaddr_in)) < 0)
	{
		perror("\nError making connection");
		exit(4);
	}

	if (telnet)
		signal(SIGINT, telnet_sigint);
	if (!silent)
	{
		struct sockaddr_in ourside;
		int addrlen = sizeof(struct sockaddr_in);

		if (getsockname(s, (struct sockaddr *) &ourside, &addrlen) < 0)
		{
			perror("Error: getsockname");
			exit(5);
		}
		printf("\nConnected. This side = %s:%u\nHit "
			   "^D to close connection, ^C to abort.\n",
			   inet_ntoa(ourside.sin_addr), ntohs(ourside.sin_port));
	}
	DoIt(s);
	exit(0);
}
